float4x4 matWorld;
float3x3 matInverseTranspose;
float4x4 matWorldViewProj;

float4 vecSunlight; // xyz must be normalized, w is for shadow?
float4 vecCameraPosition; // w is wetslider
float4 vecCameraDirectionWorld; // w is ProcessVisibleMap
float4 tDim; // x UNUSED y UNUSED w UNUSED, z is GRIME factor

//float4 colRadiosity; // not yet implemented
float4 colAmbient; // w is fog end distance
float4 colSunlight; // w is makeship-only sun on/off
float4 colInstance; // w is instance type

bool flags[7]; // 0=reflect, 1=height, 2=color, 3=overlay, 4=diffuse, 5=detailcontrol, 6=detail


Texture2D SkyboxTexture;
samplerCUBE SkyboxSampler = sampler_state
{
	texture = < SkyboxTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Mirror;
	AddressV = Mirror;
};

Texture2D ColorTexture;
sampler2D ColorSampler = sampler_state
{
	texture = < ColorTexture > ;
	magfilter = ANISOTROPIC;
	minfilter = ANISOTROPIC;
	mipfilter = ANISOTROPIC;
	MaxAnisotropy = 8;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D HeightTexture;
sampler2D HeightSampler = sampler_state
{
	texture = < HeightTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D ReflectTexture;
sampler2D ReflectSampler = sampler_state
{
	texture = < ReflectTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D DiffuseTexture;
sampler2D DiffuseSampler = sampler_state
{
	texture = < DiffuseTexture >;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D VisibleTexture;
sampler2D VisibleSampler = sampler_state
{
	texture = < VisibleTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D GrimeTexture;
sampler2D GrimeSampler = sampler_state
{
	texture = < GrimeTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D OverlayTexture;
sampler2D OverlaySampler = sampler_state
{
	texture = < OverlayTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Border;
	AddressV = Border;
	BorderColor = 0x0;
};

Texture2D DetailTexture;
sampler2D DetailSampler = sampler_state
{
	texture = < DetailTexture > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

Texture2D DetailAlpha;
sampler2D DetailAlphaSampler = sampler_state
{
	texture = < DetailAlpha > ;
	magfilter = LINEAR;
	minfilter = LINEAR;
	mipfilter = LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

struct VertexShaderInput
{
	float3 Position : POSITION0;
	float3 Normal : NORMAL0;
	float4 Color: COLOR0;
	float2 TextureColorMap: TEXCOORD0;
	float2 TextureOverlayMap: TEXCOORD1;
	float3 Tangent : TANGENT0;
	float3 Binormal : BINORMAL0;
};

struct VertexShaderOutputHigh // Also MediumToHigh
{
	float4 Position : SV_Position;
	float3 Color: COLOR0;
	float2 TextureColorMap : TEXCOORD0;
	float2 TextureOverlayMap : TEXCOORD1;
	float3 Normal: TEXCOORD2;
	float3 Tangent: TEXCOORD3;
	float3 Binormal: TEXCOORD4;
	float3 WorldPosition : TEXCOORD5;
};

//struct VertexShaderOutputMedium
//{
//	float4 Position : POSITION0;
//	float3 Color: COLOR0;
//	float3 DiffuseAndAmbient: COLOR1;
//	float2 TextureColorMap : TEXCOORD0;
//	float3 Normal: TEXCOORD1;
//};


VertexShaderOutputHigh VertexShaderFunctionHigh(VertexShaderInput input)
{
	VertexShaderOutputHigh output;

	output.Position = mul(float4(input.Position, 1), matWorldViewProj);

	output.Normal = normalize(mul(input.Normal, matInverseTranspose));
	output.Tangent = normalize(mul(input.Tangent, matInverseTranspose));
	output.Binormal = normalize(mul(input.Binormal, matInverseTranspose));

	output.Color = input.Color.rgb * colInstance.rgb;

	output.TextureColorMap = input.TextureColorMap;
	output.TextureOverlayMap = input.TextureOverlayMap;

	// these are needed for per-vertex specular 
	output.WorldPosition = mul(input.Position, matWorld).xyz;

	// can't send this guy if we want the reflection to interact with the bumpnormals
	//output.TextureReflection = 0;// reflect(vecCameraDirectionWorld.xyz, output.Normal);

	return output;
}

float4 PixelShaderFunctionHigh(VertexShaderOutputHigh input) : COLOR0
{
	float3 bumpNormal = input.Normal;
	float occlusion = 1.0f;
	if (flags[1]) // bump flag
	{
		// Sample the pixel in the bump map.
		float4 bumpMap = tex2D(HeightSampler, input.TextureColorMap);

			// Expand the range of the normal value from (0, +1) to (-1, +1). _bx2 register?
			float3 bumpMapRgb = 2 * (bumpMap.rgb - 0.5);
			occlusion = bumpMap.a;

		// Calculate the normal from the data in the bump map.
		bumpNormal = (bumpMapRgb.x * input.Tangent) + (bumpMapRgb.y * input.Binormal) + (bumpMapRgb.z * input.Normal);

		// Normalize the resulting bump normal.
		bumpNormal = normalize(bumpNormal);
	}



	float4 tColorOverlay = 0;
		if (flags[3]) // use overlay
		{
			tColorOverlay = tex2D(OverlaySampler, input.TextureOverlayMap);
		}



	// first check shininess and see if we should bother?
	float shininess = vecCameraPosition.w;
	if (flags[0] && shininess > 0) // specular texture flag
	{
		float tColor1 = tex2D(ReflectSampler, input.TextureColorMap).r;

		// only where there is no overlay
		shininess = max(shininess, (1 - tColorOverlay.a) * tColor1 * 255.0 + 1.0);
	}

	// we chose to do the vecCameraDirectionWorld as a "directional" ray instead of a "point" ray ... "point" ray would require us to
	// do a diff between each vertex location and the viewer's location, normalized as a vector
	//float dotProduct = dot(r, input.ViewDirection); // vertex based works at a distance but can't see difference between simple single
	float3 r = reflect(-vecSunlight.xyz, bumpNormal);
		float3 viewDirection = normalize(vecCameraPosition.xyz - input.WorldPosition); // per pixel
		float dotProduct = dot(r, viewDirection); // per pixel
	//float dotProduct = dot(r, vecCameraDirectionWorld.xyz); // simple single 
	float specular = 0;
	if (dotProduct > 0.0 && shininess > 0)
	{
		// was 10 when I sent it to ian, map supports 0 (disabled) and 1 (effectively disabled) to 255 (very shiny surface)
		// that translates from 1 to 100?
		//float shininess = vecCameraPosition.w;
		//if (flags[0]) // specular texture flag
		//	shininess = max(shininess, tex2D(ReflectSampler, input.TextureColorMap).r*255.0 + 1.0);

		specular = pow(abs(dotProduct), shininess); // 5 is about as low as Ian will go
		// so 255 is very shiny and 100 bright, 2 is overall not very shiny and maybe what 30% bright?
		specular *= 0.1 + shininess / 283.0;
	}


	float3 blend = input.Color;
		if (flags[2]) // use texture color map
		{
			float4 tColor = tex2D(ColorSampler, input.TextureColorMap);
				blend = (1 - tColor.a) * blend + tColor.a * tColor.rgb;
		}
	if (flags[3]) // use overlay
	{
		blend = (1 - tColorOverlay.a) * blend + tColorOverlay.a * tColorOverlay.rgb;
	}
	if (tDim.z > 0.0f) // grime, could probably modulate reflectivity as well?
	{
		blend *= lerp(1, tex2D(GrimeSampler, input.TextureColorMap).r, tDim.z);
	}
	if (flags[6]) // detail layer
	{
		// 4x frequency, midline 127 allows us to lighten or darken
		float detailMod = tex2D(DetailSampler, input.TextureColorMap * 4).r * 2;
		if (flags[5]) // detail control layer
		{
			blend *= lerp(1, detailMod, tex2D(DetailAlphaSampler, input.TextureColorMap).a);
		}
		else
		{
			blend *= detailMod;
		}
	}


	float returnalpha = 1;
	if (vecCameraDirectionWorld.w) // use alpha visibility 
		returnalpha = tex2D(VisibleSampler, input.TextureColorMap).a;


	float lightIntensity = saturate(dot(bumpNormal, vecSunlight.xyz)); // diffuse


	// diffuse and ambient has to be per-pixel
	float3 diffuseAndAmbient;
	if (colInstance.w == 38.0f) // mostly enclosed spaces
		diffuseAndAmbient = colAmbient.rgb * occlusion * 0.25f; // 25% ambient, no sun
	else
		diffuseAndAmbient = max(colSunlight.rgb * lightIntensity, colAmbient.rgb * occlusion); // diffuse and ambient

	if (flags[4]) // has diffuse layer
		diffuseAndAmbient = max(diffuseAndAmbient, tex2D(DiffuseSampler, input.TextureColorMap).r);

	// @todo could tint the specular reflection here
	float3 textureReflection = reflect(viewDirection, bumpNormal);
	float3 reflcube = texCUBE(SkyboxSampler, normalize(textureReflection)).rgb * shininess / 256 / 3;
	float3 specularAndReflection = max(specular*colSunlight.rgb, reflcube.rgb*colSunlight.w);

	// fog
	float fogfactor = saturate(length(input.WorldPosition - vecCameraPosition.xyz) / colAmbient.a);
	float3 diffuse = saturate(diffuseAndAmbient*blend + specularAndReflection);
	return float4(lerp(diffuse, float3(0,0,0), fogfactor), returnalpha);
	//return float4(lerp(diffuse, colAmbient.rgb, fogfactor), returnalpha);
}
technique Vehicles
{
	pass High0
	{
		VertexShader = compile vs_3_0 VertexShaderFunctionHigh();
		PixelShader = compile ps_3_0 PixelShaderFunctionHigh();
	}
	//pass MediumToHigh1
	//{
	//	VertexShader = compile vs_3_0 VertexShaderFunctionMediumToHigh();
	//	PixelShader = compile ps_3_0 PixelShaderFunctionMediumToHigh();
	//}
	//pass Medium2
	//{
	//	VertexShader = compile vs_3_0 VertexShaderFunctionMedium();
	//	PixelShader = compile ps_3_0 PixelShaderFunctionMedium();
	//}

	/*
	High
	Pixel reflection, Environment Map
	Pixel fog
	Bump
	Color/Diffuse/Overlay

	Medium
	Vertex refection
	Vertex fog
	No bump?
	Color/Diffuse

	Low
	Flat reflection
	Vertex fog
	No bump
	Color
	*/
}
